/* XPM */
static const char * red_xpm[] = {
"16 16 5 1",
"  c black",
". c #800000",
"X c red",
"o c white",
"O c None",
"OOOOOOOOOOOOOOOO",
"OOOOOOO.OOOOOOOO",
"OOOOOO.X OOOOOOO",
"OOOOO.oXX OOOOOO",
"OOOO.ooXXX OOOOO",
"OOO.oooXXXX OOOO",
"OO.ooooXXXXX OOO",
"O.XXXXX...... OO",
"OO.XXXX..... OOO",
"OOO.XXX.... OOOO",
"OOOO.XX... OOOOO",
"OOOOO.X.. OOOOOO",
"OOOOOO.. OOOOOOO",
"OOOOOOO OOOOOOOO",
"OOOOOOOOOOOOOOOO",
"OOOOOOOOOOOOOOOO"
};
